import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.awt.geom.*;
import java.math.*;
import java.awt.image.*;
import java.io.File;
import javax.imageio.ImageIO;

public class PictureCanvas extends ScaleCanvas implements MouseListener,MouseMotionListener,KeyListener {
    Manager M;
    BufferedImage IMAGE;
    Point JX;
    QI2Polygon DOMAIN,RANGE;
    Complex[] POINT=new Complex[10000];
    int POINTCOUNT;
    Tracer TR;
    PolyVector V;
    int INDEX;
    boolean DRAG;

     public PictureCanvas() {
	 addMouseListener(this);
	 addMouseMotionListener(this);
	 addKeyListener(this);
         IMAGE = new BufferedImage(600,600,BufferedImage.TYPE_INT_RGB);
	 setScales(188,408,20);
	 SOURCE=new Complex(5.66,1.68);
	 initializePolygons();
	 TR=new Tracer();
	 POINTCOUNT=0;
	 JX=new Point(0,0);
	 DRAG=false;
     }

    public void initializePolygons() {
	DOMAIN=new QI2Polygon();
	RANGE=new QI2Polygon();
	RANGE.COUNT=4;
	for(int i=0;i<10;++i) {
	    RANGE.V[i][0]=new QI2(0,0);
	    RANGE.V[i][1]=new QI2(0,0);
	    DOMAIN.V[i][0]=new QI2(0,0);
	    DOMAIN.V[i][1]=new QI2(0,0);
	}
    }


   public void paint(Graphics g2) {
      Graphics2D g=(Graphics2D) g2;
      g.drawImage(IMAGE,0,0,null);
      drawDisplay(g);
   }  


    public void drawDisplay(Graphics2D g) {
	int N=6;
	int[] k=new int[N];
	for(int i=0;i<N;++i) k[i]=M.C.DISPLAY.L[i].on;
	Color[] c=new Color[N];
	for(int i=0;i<N;++i) c[i]=M.C.DISPLAY.M[i].C;
	if(k[0]==1) drawFrame(g,c[0]);	
	if(M.C.MC.G[3].val==1) drawUnitSquare(g);
	if(k[3]==1) {
           RenderPartition.tri(M,g,c[3]);
	   RenderPartition.triSelect(M,g,c[4]);
	}
	if(k[4]==1) drawPolygon(g);
	if(k[5]==1) DemoGauss.main(this.M,g);  
        if(k[2]==1) RenderCore.core(M,g);
	if(k[1]==1) drawPointer(g,c[1]);
	TR.render(this,g);
	drawCoords(g,c[0]);
	if(k[3]==1) drawPiece(g);	
    }


    /**Draws the frame*/

    public void drawFrame(Graphics2D g,Color C) {
	g.setColor(Color.white);
	g.drawRect(0,0,getWidth()-1,getHeight()-1);
	Path2D.Double gp=new Path2D.Double();
	double phi=(1+Math.sqrt(5))/2;
	double x=1/(phi*phi);
	double x1=QI.innerBound();
	double y1=QI.innerBound();
	double x2=QI.outerBound();
	double y2=QI.outerBound();
	double y3=-y2*QI.phi()*QI.phi();

	gp.moveTo(x1,y2);
	gp.lineTo(x2,y3);
	gp.moveTo(y2,x1);
	gp.lineTo(y3,x2);
	gp.moveTo(x1,y1);
	gp.lineTo(x2,y2);

	gp.moveTo(x1,y1);
	gp.lineTo(x1,y2);
	gp.lineTo(x2,y2);
	gp.lineTo(x2,y1);
	gp.closePath();

	gp=transform(gp);
	g.setColor(C);
	g.draw(gp);  
    }

    /**Draws the coords*/

    public void drawCoords(Graphics2D g,Color C) {
	g.setColor(Color.black);
	g.fillRect(0,680,340,20);
	g.setColor(Color.white);
	g.drawRect(0,680,340,20);
        g.setFont(new Font("Helvetica",Font.PLAIN,14));
        g.setColor(Color.white);
        Complex z=new Complex(SOURCE);
        if(SpecialAffine.GO==true) z=SpecialAffine.map(z);
	String S1=GraphicsHelp.cleanLong(z.x);
	String S2=GraphicsHelp.cleanLong(z.y);
        g.drawString(S1+"    "+S2,5,getHeight()-5);
    }

    public void drawPiece(Graphics2D g) {
        g.setFont(new Font("Helvetica",Font.PLAIN,20));
	int n=SelectPartition.TILE;
	String S=TileCanvas.recognizeTile(n);
	g.setColor(Color.white);
	g.drawString(S,5,20);
    }


    /**unit square*/

    public void drawUnitSquare(Graphics2D g) {
	Path2D.Double gp=new Path2D.Double();
	gp.moveTo(0,0);
	gp.lineTo(1,0);
	gp.lineTo(1,1);
	gp.lineTo(0,1);
	gp.closePath();
	gp=transform(gp);
	g.setColor(Color.black);
	g.fill(gp);
	g.setColor(Color.white);
	g.draw(gp);
    }



    /**draws segment between point and image of point
       under the map.*/

    public void drawPointer(Graphics2D g,Color C) {
	int[] a=M.C.MC.getMap();
	Path2D.Double gp=new Path2D.Double();
	double[] d1={SOURCE.x,SOURCE.y};
	double[] d2={0,0};
	d2=HeatSemigroup.main(a,d1);
	gp.moveTo(d1[0],d1[1]);
	gp.lineTo(d2[0],d2[1]);
	g.setStroke(new BasicStroke(2));
	gp=transform(gp);
	g.setColor(C);
	g.draw(gp);
	g.setStroke(new BasicStroke(1));
    }


    /**This draws the polygon associated to the poly feature*/



    public void drawPolygon(Graphics2D g) {

	Color[] C=new Color[6];
	int[] on=new int[6];
	for(int i=0;i<6;++i) {
	    on[i]=M.C.POLY.L[i].on;
            C[i]=M.C.POLY.M[i].C;
	}

	PolyVector Q=new PolyVector();
	Path2D.Double gp=new Path2D.Double();
	if(DOMAIN==null) return;
	if(DOMAIN.COUNT==0) return;
	Q=DOMAIN.toDouble();
	gp=Q.toPath();
	gp=transform(gp);

	/**domain fill*/
	if(on[0]==1) {
 	  g.setColor(C[0]);
	  g.fill(gp);
	}

	/**domain outline*/
	if(on[1]==1) {
 	  g.setColor(C[1]);
	  g.draw(gp);	
   	  int N=SelectStyle.NUMBER1;
	  GraphicsHelp.label(g,gp,N);
	}

	/**range fill*/
	if(on[2]==1) {
	   int[] A=M.C.MC.getMap();
	   int LIM=PlotStyle.power(4,PlotStyle.DENS);
           GraphicsHelp.fillMap(this,g,A,Q,LIM,C[2],2);
	}

	/**range outline*/
	if(on[3]==1) {
	   int[] A=M.C.MC.getMap();
	   int LIM=PlotStyle.power(4,PlotStyle.DENS);
           GraphicsHelp.drawMap(this,g,A,Q,LIM,C[3],2);
	}

	if(RANGE==null) return;
	gp=new Path2D.Double();
	Q=RANGE.toDouble();
	gp=Q.toPath();
	gp=transform(gp);

	/**container fill*/
	if(on[4]==1) {
 	  g.setColor(C[4]);
	  g.fill(gp);
	}

	/**container outline*/
	if(on[5]==1) {
	  g.setColor(C[5]);
	  g.draw(gp); 
          int N=SelectStyle.NUMBER2;
	  GraphicsHelp.label(g,gp,N);
	}

    }




    /**This sets the polygon*/

    public void setPolygon(int mode,Point X) {
	int q=SelectStyle.WHICH;
	int r=SelectStyle.LOCK;	
        QI2[] p=processPoint(X);

	if(q==0) {
	  DOMAIN.COUNT=SelectStyle.NUMBER1;
	  if(r!=1)  DOMAIN.V[mode][0]=new QI2(p[0]);
	  if(r!=2)  DOMAIN.V[mode][1]=new QI2(p[1]);
	}
	if(q==1) {
	  RANGE.COUNT=SelectStyle.NUMBER2;
	  if(r!=1)  RANGE.V[mode][0]=new QI2(p[0]);
	  if(r!=2)  RANGE.V[mode][1]=new QI2(p[1]);
	}
    }

    /**This cycles the vertices*/
    public void cycleVertices(int i) {
	int q=SelectStyle.WHICH;
	if(q==1) RANGE=RANGE.rotate(i);
	if(q==0) DOMAIN=DOMAIN.rotate(i);
    }

    /**processing points*/
    public QI2[] processPoint(Point X) {
	int t=SelectStyle.MODE;
	if(t==0) return(processPointClick(X));
	return(processPointKey(X));
    }

    /**processing points*/
    public QI2[] processPointKey(Point X) {
	int[] a=new int[4];

	for(int i=0;i<4;++i) a[i]=SelectStyle.N[i];
	QI x0=new QI(a[0],a[1]);
	QI y0=new QI(a[2],a[3]);
	QI2[] zz={new QI2(x0),new QI2(y0)};
	return(zz);
    }


    /**processing points*/
    public QI2[] processPointClick(Point X) {
	Complex z=unTransform(X);
	int k=SelectStyle.COERCE;
	double[] list={100,.01,.002,.0005,.0001};
	double d=list[k];
	int index=DataPartitionManager.index(z.x,z.y);
	double[] p=DataPartitionManager.pointN(index);
	Complex w=new Complex(p[0],p[1]);
	if(Complex.dist(z,w)<d) {
	    return(DataPartitionManager.point(index));
	}

	QI x0=QI.approx(z.x,d);
	QI y0=QI.approx(z.y,d);
	QI2[] zz={new QI2(x0),new QI2(y0)};
	return(zz);

    }



    /**Mouse events*/

    public void mouseDragged(MouseEvent e) { 
	MouseData J=MouseData.process(e);
	if(J.mode==2) {
            SOURCE=unTransform(J.X);
	}
    }

    public void mouseMoved(MouseEvent e) { 
	MouseData J=MouseData.process(e);
	JX=J.X;
	if(DRAG==true) 	{
          SOURCE=unTransform(JX);
	  M.repaint();
	}
    }

    public void mousePressed(MouseEvent e) { 
    }


     public void mouseReleased(MouseEvent e) {	 
     }

     public void mouseEntered(MouseEvent e) {
	 requestFocus();
     }

     public void mouseExited(MouseEvent e) {}   

     public void mouseClicked(MouseEvent e) {
	MouseData J=MouseData.process(e);
	doMouseClick(J.mode);
     }


     public void doMouseClick(int mode) {
	if(mode==1) scaleUp(JX,-1);
	if(mode==3) scaleUp(JX,+1);
	if(mode==2) SOURCE=unTransform(JX);
	if(mode==2) RenderPartition.selectCorrectMap(M);
	if((mode>10)&&(mode<100)) setPolygon(mode-11,JX);
	if(mode==100) cycleVertices(-1);
	if(mode==101) cycleVertices(+1);
	if((mode==1)||(mode==3)) M.C.doRun();
	M.repaint();
     }


    public void keyPressed(KeyEvent e) {
	int test=-100;
	DRAG=true;
	char ch=e.getKeyChar();
	if(ch=='z') test=1;   //zoom in
	if(ch=='x') test=2;
	if(ch=='c') test=3;   //zoom out
	if(ch=='1') test=11;   //vertex select1
	if(ch=='2') test=12;   //vertex select2
	if(ch=='3') test=13;   //vertex select3
	if(ch=='4') test=14;   //vertex select4
	if(ch=='5') test=15;   //vertex select5
	if(ch=='6') test=16;   //vertex select6
	if(ch=='7') test=17;   //vertex select7
	if(ch=='8') test=18;   //vertex select8
	if(ch=='9') test=19;   //vertex select9
	if(ch=='a') test=20;   //vertex select10
	if(ch=='b') test=21;   //vertex select11
	if(ch=='-') test=100;  //cycle vertices -
	if(ch=='=') test=101;  //cycle vertices +
        doMouseClick(test);
	M.repaint();
    }

    public void keyTyped(KeyEvent e) {}
    public void keyReleased(KeyEvent e) {
	DRAG=false;
    }


    /**Exporting routine*/

    public void doExport() {
        File f = new File("test.png");
        try{ ImageIO.write(IMAGE, "PNG", f);}
	catch (Throwable t) {}
    }

}
