import java.applet.Applet;
import java.awt.event.*;
import java.awt.*;

/**These are 2-variable polynomials, up to maxdegree 19,
   with coefficients in Z[sqrt 5].*/

public class QI2Polynomial {
    public static int SIZE=16;
    QI2[][] c=new QI2[SIZE][SIZE];


    /**returns the 0 polynomial*/
    public QI2Polynomial() {
	for(int i=0;i<SIZE;++i) {
	    for(int j=0;j<SIZE;++j) {
		this.c[i][j]=QI2.zero();
	    }
	}
    }	

    /**returns the polynomial which is just a constant r*/
    public QI2Polynomial(QI2 r) {
	for(int i=0;i<SIZE;++i) {
	    for(int j=0;j<SIZE;++j) {
		this.c[i][j]=QI2.zero();
	    }
	}
        c[0][0]=new QI2(r);
    }

    /**returns the linear polynomial Ax+By+C*/

    public QI2Polynomial(QI2 A,QI2 B,QI2 C) {
	for(int i=0;i<SIZE;++i) {
	    for(int j=0;j<SIZE;++j) {
		this.c[i][j]=QI2.zero();
	    }
	}
        c[1][0]=new QI2(A);
        c[0][1]=new QI2(B);
        c[0][0]=new QI2(C);
    }

    /**copies a polynomial*/

    public QI2Polynomial(QI2Polynomial P) {
	for(int i=0;i<SIZE;++i) {
	    for(int j=0;j<SIZE;++j) {
		this.c[i][j]=new QI2(P.c[i][j]);
	    }
	}
    }

    /**for our purposes we just need the case when
        A and B are square arrays of the same size.*/

    public QI2Polynomial(long[][] A,long[][] B) {
	for(int i=0;i<SIZE;++i) {
	    for(int j=0;j<SIZE;++j) {
		c[i][j]=QI2.zero();
	    }
	}
	for(int i=0;i<A.length;++i) {
	    for(int j=0;j<A.length;++j) {
		c[i][j]=new QI2(A[i][j],B[i][j]);
	    }
	}
    }


    /**scaling*/
    public QI2Polynomial scale(QI2 r) {
	return(times(this,new QI2Polynomial(r)));
    }


    /**addition*/
    public static QI2Polynomial plus(QI2Polynomial X,QI2Polynomial Y) {
	QI2Polynomial Z=new QI2Polynomial();
	for(int i=0;i<SIZE;++i) {
	    for(int j=0;j<SIZE;++j) {
		Z.c[i][j]=QI2.plus(X.c[i][j],Y.c[i][j]);
	    }
	}
	return(Z);
    }


    /**subtraction*/
    public static QI2Polynomial minus(QI2Polynomial X,QI2Polynomial Y) {
	QI2Polynomial Z=new QI2Polynomial();
	for(int i=0;i<SIZE;++i) {
	    for(int j=0;j<SIZE;++j) {
		Z.c[i][j]=QI2.minus(X.c[i][j],Y.c[i][j]);
	    }
	}
	return(Z);
    }

    /**multiplication. this will only work if the
       degree of the product is less than SIZE.*/


    public static QI2Polynomial times(QI2Polynomial X,QI2Polynomial Y) {
	QI2Polynomial Z=new QI2Polynomial();
	int[] b1=X.getBound();
	int[] b2=Y.getBound();

	for(int i1=0;i1<=b1[0];++i1) {
	    for(int j1=0;j1<=b1[1];++j1) {
  	       for(int i2=0;i2<=b2[0];++i2) {
	           for(int j2=0;j2<=b2[1];++j2) {
		       boolean test1=X.c[i1][j1].isZero();
		       boolean test2=Y.c[i2][j2].isZero();
		       if((test1==false)&&(test2==false)) {
		         int i=i1+i2; 
			 int j=j1+j2;
			 QI2 temp=QI2.times(X.c[i1][j1],Y.c[i2][j2]);
		         Z.c[i][j]=QI2.plus(Z.c[i][j],temp);
		       }
		   }
	       }
	    }
	}
	return(Z);
    }

    /**this only works if the final degree is less than SIZE.  We only
       use this routine when x is a linear functional and j<SIZE.**/

    public static QI2Polynomial pow(QI2Polynomial x,int j) {
	QI2Polynomial y=new QI2Polynomial(QI2.one());
	for(int i=0;i<j;++i) {
	    y=times(y,x);
	}
	return(y);
    }

    /**negation: replaces P with -P*/

    public QI2Polynomial negate() {
	QI2Polynomial Z=new QI2Polynomial();
	for(int i=0;i<SIZE;++i) {
	    for(int j=0;j<SIZE;++j) {
		Z.c[i][j]=c[i][j].negate();
	    }
	}
	return(Z);
    }

    /**composes with the affine transformation
       T(x,y)= (a10 x + a01 y + a00,b10 x + b01 y + b00)*/

    public QI2Polynomial composeWithAffine(QI2 A10,QI2 A01,QI2 A00,QI2 B10,QI2 B01,QI2 B00) {
	QI2Polynomial X=new QI2Polynomial();
	QI2Polynomial a=new QI2Polynomial(A10,A01,A00);
	QI2Polynomial b=new QI2Polynomial(B10,B01,B00);
	QI2Polynomial m1=new QI2Polynomial();
	QI2Polynomial m2=new QI2Polynomial();
	QI2Polynomial m3=new QI2Polynomial();
	int[] bb=getBound();
	for(int i=0;i<bb[0];++i) {
	    for(int j=0;j<bb[1];++j) {
		if(c[i][j].isZero()==false) {
		    m1=pow(a,i);
		    m2=pow(b,j);
		    m3=times(m1,m2);
		    m3=m3.scale(c[i][j]);
		    X=plus(X,m3);
		}
	    }
	}
	return(X);
    }

    /**Composes with the affine transformation which has the action
       (0,1]) -> (A[0],A[1])
       (1,0) -> (B[0],B[1])
       (0,1) -> (C[0],C[1])
       So, the value of the output poly on the standard unit triangle is
       the same as the value of the input poly on he triangle with vertices A,B,C
    **/

    public QI2Polynomial moveTriangle(QI2Polygon V) {
	return(moveTriangle(V.V[0],V.V[1],V.V[2]));
    }

    public QI2Polynomial moveTriangle(QI2[] A,QI2[] B,QI2[] C) {
	QI2 A00=new QI2(A[0]);
	QI2 A10=QI2.minus(B[0],A[0]);
	QI2 A01=QI2.minus(C[0],B[0]);
	QI2 B00=new QI2(A[1]);
	QI2 B10=QI2.minus(B[1],A[1]);
	QI2 B01=QI2.minus(C[1],B[1]);
	return(this.composeWithAffine(A10,A01,A00,B10,B01,B00));
    }



    /**forms the new polynomial Q(x,y)=P(x,xy)*/

    public QI2Polynomial squareTriangleTransform() {
	QI2Polynomial Q=new QI2Polynomial();
	int[] b=getBound();
	for(int i=0;i<b[0];++i) {
	    for(int j=0;j<b[1];++j) {
		if(c[i][j].isZero()==false) {
		   Q.c[i+j][j]=new QI2(c[i][j]);
		}
	    }
	}
	return(Q);
    }

    public int firstTermSign() {
	QI2 d=firstTerm();
	return(d.sign());
    }

    public QI2 firstTerm() {
	int[] b=getBound();
	for(int i=0;i<b[0];++i) {
	    for(int j=0;j<b[1];++j) {
		int s=c[i][j].sign();
		if(s!=0) return(c[i][j]);
	    }
	}
	return(new QI2(0,0));
    }


    /**main positive dominance tests*/
    public int checkDominance(QI2Polygon W,int strict) {
       QI2Polynomial Z=this.moveTriangle(W);
       Z=Z.squareTriangleTransform(); 
       int sign=Z.dominanceTest(strict);
       return(sign);
    }

    /**positive dominance*/

    public int dominanceTest(int strict) {
	if(strict==0) return(dominanceTestSoft());
	return(dominanceTestHard());
    }

    public int dominanceTestSoft() {
         int sign=firstTermSign();
	 if(sign>0) {
	     boolean test=isPositiveDominantSoft();
	     if(test==true) return(1);
	 }
	 if(sign<0) {
             boolean test=isNegativeDominantSoft();
	     if(test==true) return(-1);
	 }
	 return(0);
    }

    public int dominanceTestHard() {
         int sign=firstTermSign();
	 if(sign>0) {
	     boolean test=isPositiveDominantHard();
	     if(test==true) return(1);
	 }
	 if(sign<0) {
             boolean test=isNegativeDominantHard();
	     if(test==true) return(-1);
	 }
	 return(0);
    }





    /**Checks if the polynomial is positive on 3 points on the
       edges of the unit square.  We ignore the vertical edge
       through the origin. This is a technical check
       used in the positive dominance test.**/

    public boolean edgeCheck() {
	QI2[] p2={new QI2(1,0),new QI2(-1,1)};
	QI2[] p3={new QI2(-1,1),new QI2(0,0)};
	QI2[] p4={new QI2(-1,1),new QI2(1,0)};
	QI2[][] p={p2,p3,p4};
	for(int i=0;i<3;++i) {
	    QI2 test=this.evaluate(p[i]);
	    if(test.isZero()==true) return(false);
	}
	return(true);
    }

    /**Polynomial is non-negative on the unit square
       and positive in the interior*/

    public boolean isPositiveDominantSoft() {
	int[] b=getBound();
	for(int i=0;i<=b[0];++i) {
	    for(int j=0;j<=b[1];++j) {
		boolean test=positiveOrZeroSum(i,j);
		if(test==false) return(false);
	    }
	}
	boolean test=positiveSum(b[0],b[1]);
	if(test==false) {
	    return(false);
	}
	return(true);
    }



    /**Polynomial is POSITIVE on the unit square*/
    public boolean isPositiveDominantHard() {
	int[] b=getBound();
	for(int i=0;i<=b[0];++i) {
	    for(int j=0;j<=b[1];++j) {
		boolean test=positiveSum(i,j);
		if(test==false) return(false);
	    }
	}
	return(true);
    }

    /**Checks if the sum of the coefficients up to (a,b) is positive or zero*/
    public boolean positiveOrZeroSum(int a,int b) {
	QI2 C=sum(a,b);
	return(C.isPositiveOrZero());
    }


    /**Checks if the sum of the coefficients up to (a,b) is positive*/
    public boolean positiveSum(int a,int b) {
	QI2 C=sum(a,b);
	return(C.isPositive());
    }


    /**Checks if the sum of the coefficients up to (a,b) is positive*/
    public QI2 sum(int a,int b) {
	QI2 C=new QI2(new QI(0,0),new QI(0,0),0);
	for(int i=0;i<=a;++i) {
	    for(int j=0;j<=b;++j) {
		C=QI2.plus(C,c[i][j]);
	    }
	}
	return(C);
    }


    /**negative dominant*/
    public boolean isNegativeDominantSoft() {
	QI2Polynomial R=this.negate();
	return(R.isPositiveDominantSoft());
    }

    /**negative dominant*/
    public boolean isNegativeDominantHard() {
	QI2Polynomial R=this.negate();
	return(R.isPositiveDominantHard());
    }

    /**exact evaluation*/

    public QI2 evaluate(QI2[] X) {
	return(evaluate(X[0],X[1]));
    }


    public QI2 evaluate(QI2 X,QI2 Y) {
	QI2 T=QI2.zero();
	int[] bound=getBound();
	for(int i=0;i<bound[0];++i) {
	    for(int j=0;j<bound[1];++j) {
		QI2 A=QI2.pow(X,i);
		QI2 B=QI2.pow(Y,j);
		T=QI2.plus(T,QI2.times(c[i][j],QI2.times(A,B)));
	    }
	}
	return(T);
    }

    /**multiplies the polynomial by +/- 1 so that it is non-negative
       on the given point.*/

    public QI2Polynomial normalize(QI2[] X) {
	QI2 Y=this.evaluate(X);
	if(Y.isPositive()==true) return(this);
	return(this.negate());
    }




    /**exact evaluation of signs.*/

    public static int evaluateSign(QI2RationalFunction R,QI2[] X) {
	QI2 p=R.P.evaluate(X);
	QI2 q=R.Q.evaluate(X);
	int sp=p.sign();
	int sq=q.sign();
	if(sp==99) return(99);
	if(sq==99) return(99);
	if(sq==0) return(99);
	if(sp==sq) return(1);
	return(-1);
    }


    /**numerically evaluates on a pair of doubles.  This is
       used for debugging purposes.*/

    public double evaluate(double x,double y) {
	double t=0;
	for(int i=0;i<SIZE;++i) {
	    for(int j=0;j<SIZE;++j) {
		if(c[i][j].isZero()==false) {
		  double a=c[i][j].toDouble();
                  t=t+a*Math.pow(x,i)*Math.pow(y,j);
		}
	    }
	}
	return(t);
    }

    /**finds the highest degrees with nonzero coefficients.
       This makes the printout nicer.*/

    public int[] getBound() {
	int a=0;
	int b=0;
	for(int i=0;i<SIZE;++i) {
	    for(int j=0;j<SIZE;++j) {
		if(c[i][j].isZero()==false) {
		    if(a<i) a=i;
		    if(b<j) b=j;
		}
	    }
	}
	int[] A={a+1,b+1};
	return(A);
    }


    /**printing routines*/

    public void print() {
	int[] s=this.getBound();
	System.out.println("---------------------");
	for(int i=0;i<s[0];++i) {
	    for(int j=0;j<s[1];++j) {
		if(c[i][j].isZero()==false) {
		  System.out.print("x^"+i);
		  System.out.print(" y^"+j+"  "); 
                  c[i][j].println();
		}
	    }
	}
    }

    public void printN() {
	printN(SIZE);
    }

    public void printN(int mm) {
	QI2 d0=firstTerm();
	if(d0.isZero()==true) return;
	double d=d0.toDouble();
	boolean test=false;
	int m=SIZE;
	if(m>mm) m=mm;
	for(int k=0;k<m;++k) {
	  for(int i=0;i<=k;++i) {
	      int j=k-i;
		if((i<m)&&(j<m)) {
		    if(c[i][j].isZero()==false) {
		        test=true;
		        System.out.print("(");
		        double b=c[i][j].toDouble();
		        System.out.print(b/d);
		        System.out.print(") ");
		        if(i==1) System.out.print("x");
		        if(j==1) System.out.print("y");
		        if(i>1) System.out.print("x^"+i);
		        if(j>1) System.out.print("y^"+j);
		        System.out.println(" + ");
		    }
		}
	  }
	}
	System.out.println("---------------------");
    }

    /**rendering routine. used with the algebra display.*/
    public void render(Graphics2D g,int X,int Y,Color C) {
	int posX=X;
	int posY=Y;    
        g.setFont(new Font("Helvetica",Font.PLAIN,12));
	g.setColor(C);
	for(int i=0;i<SIZE;++i) {
	    for(int j=0;j<SIZE;++j) {
		if(c[i][j].isZero()==false) {
		  QI temp=new QI(c[i][j].a);
		  String S="";
		  if(temp.a.longValue()>0) S=S+"+";
		  if(temp.a.longValue()!=0) {
		    S=S+temp.a.toString();
		  }
		  if(temp.b.longValue()>0) 	  S=S+"+";
		  if(temp.b.longValue()!=0) {
		    S=S+temp.b.toString();
		    S=S+" \u03D5";
		  }
		  String T="";
		  T=T+"   ";
		  Integer I=new Integer(i);
         	  Integer J=new Integer(j);    
                  if(i==1) T=T+"x";
                  if(i>1) T=T+"x^"+I.toString();  
                  if(j==1) T=T+"y";  
                  if(j>1) T=T+"y^"+J.toString();
		  g.drawString(S,posX,posY);
		  g.drawString(T,posX+100,posY);
		  posY=posY+20;
		}
	    }
	}
    }
}





